#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: bootstrapagent.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for bootstrapping agent.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

NOTE: The BootstrapAgent class is responsible to purge the stream (ie. index in stream is at COMMAND_PROMPT)

"""

from socket import timeout
import os.path
from urllib.error import URLError
import urllib.request
from time import sleep
import json
import pexpect
from voluptuous import Schema, Required, Any, MultipleInvalid
from .logging import LoggerMixin

__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''10-05-2019'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# Constant for Pexpect. This prompt is default for Fedora and CentOS.
COMMAND_PROMPT = '[#$] '

# payload
HTTP_RESPONSE = Schema({Required("result"): Any(True, False)})


class BootstrapAgent(LoggerMixin):
    """Responsible for executing the python agent module."""

    def __init__(self, tunnel, group_ports, deploy_path):
        super().__init__()
        self.tunnel = tunnel
        self.local_port_agent = group_ports["local_port_agent"]
        self.host = '127.0.0.1'  # localhost, client
        self.deploy_path = deploy_path
        self.path_to_agent = os.path.join(self.deploy_path, 'agent.py')

    def __str__(self):
        return 'BootstrapAgent'

    def start(self):
        """Executes the agent python module (after checking for availability and running process)."""
        result = False
        if self.is_agent_module_available():
            if self._can_agent_be_probed():
                if not self._instruct_stop_agent():
                    self._logger.warning('running Agent could not be instructed to stop')
                    if self._is_process_alive():
                        self._killing_process()
                        self._logger.warning('running Agent process has been killed')
            else:
                if self._is_process_alive():
                    self._killing_process()
                    self._logger.warning('running Agent process has been killed')
            result = self._execute_agent()
        return result

    def _can_agent_be_probed(self):
        """Determines whether the agent is listening thus active on target destination host."""
        result = False
        self._logger.debug('determining if Agent is alive by sending a GET request')
        try:
            with urllib.request.urlopen(f'http://{self.host}:{self.local_port_agent}',
                                        timeout=3, data=None) as request_obj:
                http_code = request_obj.getcode()
                if http_code == 200:
                    self._logger.debug('Agent is alive')
                    result = True
        except ConnectionResetError as exp:
            self._logger.debug('sending request raised an error: %s', exp)
        except URLError as exp:
            self._logger.debug('sending request raised an error: %s', exp)
        except timeout:
            self._logger.debug('timeout exceeded when instructing the Agent')
        return result

    def _remove_remnants(self):
        command = f"rm -f {self.deploy_path}agent.pid"
        self._logger.debug('removing pid file, if present: %s', command)
        self.tunnel.child.sendline(command)
        self.tunnel.child.expect(COMMAND_PROMPT)

    def _is_process_alive(self):
        """Determines if an Agent process is (still) running."""
        result = False
        # command = f"pgrep -fc '/bin/python3.6 {self.path_to_agent}'"
        command = f'if [ `pgrep -fc \'/bin/python3.6 {self.path_to_agent}\'` == 0 ]; ' \
                  f'then echo "count=0"; else echo "count=1"; fi'
        self._logger.debug('determining if an Agent process is (still) running: %s', command)
        self.tunnel.child.sendline(command)
        # self.tunnel.child.logfile = sys.stdout
        index = self.tunnel.child.expect(['count=0', 'count=1', pexpect.TIMEOUT], timeout=3)
        self.tunnel.child.expect(COMMAND_PROMPT)
        if index == 0:
            self._logger.debug('no Agent process is running')
            self._remove_remnants()
        if index == 1:
            self._logger.debug('an Agent process is (still) running on destination host')
            result = True
        return result

    def _killing_process(self):
        """Kills the Agent process."""
        command = f"pkill -f '/bin/python3.6 {self.path_to_agent}'"
        self._logger.debug('killing running agent.py process: %s', command)
        self.tunnel.child.sendline(command)
        self.tunnel.child.expect(COMMAND_PROMPT)
        while self._is_process_alive():
            self._logger.debug('waiting for OS to successfully kill process... ')
            sleep(4)
        sleep(1)
        return True

    def _instruct_stop_agent(self):
        """Stops the Agent by sending a 'process:stop' request."""
        result = False
        json_instruction = json.dumps({'process': 'stop',
                                       'arguments': {}})  # serialize dict to a JSON formatted str
        data = json_instruction.encode('utf-8')  # encode JSON formatted string to byte
        try:
            with urllib.request.urlopen(f'http://{self.host}:{self.local_port_agent}',
                                        timeout=3, data=data) as request_obj:
                response_string = request_obj.read().decode('utf-8')  # from byte to string (JSON format)
            response_dict = json.loads(response_string)  # from JSON to dict
            response = HTTP_RESPONSE(response_dict)  # validating the structure of the content of an HTTP request
            result = response.get('result')
            if result:
                self._logger.debug('Agent has received instruction to stop')
                result = True
        except URLError as exp:  # urllib.request.urlopen()
            self._logger.debug('instructing the Agent raised an error: %s', exp)
        except ConnectionResetError as exp:  # urllib.request.urlopen()
            self._logger.debug('instructing the Agent raised an error: %s', exp)
        except timeout:  # urllib.request.urlopen()
            self._logger.debug('timeout exceeded when instructing the Agent')
        except json.decoder.JSONDecodeError:  # json.loads()
            self._logger.error('response of Agent could not be read: JSON document could not be deserialized')
        except MultipleInvalid:  # HTTP_RESPONSE()
            self._logger.error('response of Agent could not be read: data structure validating failed')
        sleep(2)  # give the Agent time to terminate servers
        return result

    def is_agent_module_available(self):
        """Determines if agent.py module is on target destination host."""
        self.tunnel.child.sendline(f'file {self.path_to_agent}')
        index = self.tunnel.child.expect(['Python script', 'cannot open'])
        if index == 0:
            self._logger.debug('module agent.py is available on host')
            result = True
        else:
            self._logger.error('module agent.py is not available on host')
            result = False
        self.tunnel.child.expect(COMMAND_PROMPT)
        return result

    def _execute_agent(self):
        """Executes agent.py module on target destination host."""
        command = f'/bin/python3.6 {self.path_to_agent} {self.deploy_path}'
        self._logger.debug('executing the Agent module with command: %s', command)
        self.tunnel.child.sendline(command)
        index = self.tunnel.child.expect([COMMAND_PROMPT, 'SyntaxError', 'ModuleNotFoundError', 'AttributeError',
                                          'pid exists', pexpect.TIMEOUT], timeout=3)
        if index == 0:
            result = True
            sleep(2)  # The Agent needs some time to initialize, so wait with returning the Boolean result
        elif index == 1:
            self._logger.error(
                'check if Python version 3.6 is installed on destination host. '
                'the command was: %s', command)
            result = False
        elif index == 4:
            self._logger.error(
                'agent.py could not be executed. '
                'it seems the Agent is running due to the existence of the PID file.'
                'the command was: %s', command)
            result = False
        else:
            self._logger.error(
                'agent.py could not be executed. '
                'try running %s on destination host manually to determine the cause',
                command)
            result = False
        return result

    def remove(self):
        """Removes the python Agent module (not implemented, yet)."""
        # pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: miscellaneous.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for miscellaneous.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""
import json
import logging
import logging.config
import threading
import subprocess
from urllib.error import URLError
import urllib.request
from time import sleep
from socket import timeout
from voluptuous import Schema, Required, Any, REMOVE_EXTRA, Optional, MultipleInvalid
from .powermolelibexceptions import InvalidConfigurationFile
from .logging import LOGGER_BASENAME as ROOT_LOGGER_BASENAME, LoggerMixin


__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''10-05-2019'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')
LOGGER = logging.getLogger(f'{ROOT_LOGGER_BASENAME}.miscellaneous')  # non-class objects like fn will consult this object

# Constants
HEARTBEAT_DURATION = 5

MODE_SCHEMA = Schema({Required("mode"): Any("TOR", "FOR", "PLAIN"),  # missing mode will raise "MultipleInvalid"
                      }, extra=REMOVE_EXTRA)

TOR_SCHEMA = Schema({"mode": "TOR",
                     "gateways": [{"host_ip": str,
                                   "user": str,
                                   "identity_file": str}],
                     "destination": {"host_ip": str,
                                     "user": str,
                                     "identity_file": str},
                     Optional("application"): {"binary_name": str,
                                               "binary_location": str}
                     }, required=True)

FOR_SCHEMA = Schema({"mode": "FOR",
                     "gateways": [{"host_ip": str,
                                   "user": str,
                                   "identity_file": str}],
                     "destination": {"host_ip": str,
                                     "user": str,
                                     "identity_file": str},
                     "forwarders": [{"local_port": int,
                                     "remote_interface": str,
                                     "remote_port": int}],
                     Optional("application"): {"binary_name": str,
                                               "binary_location": str}
                     }, required=True)

PLAIN_SCHEMA = Schema({"mode": "PLAIN",
                       "gateways": [{"host_ip": str,
                                     "user": str,
                                     "identity_file": str}],
                       "destination": {"host_ip": str,
                                       "user": str,
                                       "identity_file": str},
                       Optional("application"): {"binary_name": str,
                                                 "binary_location": str}
                       }, required=True)


class Configuration(LoggerMixin):  # pylint: disable=too-few-public-methods
    """Parses the configuration file and composes the local forwarding string (SSH -L)."""

    def __init__(self, config_file):
        """Initializes the Configuration object.

        Args:
            config_file(basestring): A file containing a JSON document.

        """
        super().__init__()
        try:
            config = self.get_config(config_file)
            self.mode = config.get('mode')
            self.gateways = config.get('gateways')
            # ex. [{'host_ip': '10.10.1.72', 'user': 'root', 'identity_file': '~/.ssh/id_rsa_be_vm'}]
            self.application = config.get('application')
            # ex. _________________________
            self.destination = config.get('destination')
            # ex. {'host_ip': '10.10.2.92', 'user':'root', 'identity_file': '~/.ssh/id_rsa_it_vm'}
            forwarders = config.get('forwarders')
            # ex. [{'local_port': 587, 'remote_port': 587}, {'local_port': 995, 'remote_port': 995}]
            self.forwarders_string = ' '.join([f'-L:{forwarder["local_port"]}:'
                                               # IPv6 addresses should be enclosed in single quotes on macOS'
                                               f'\'{forwarder["remote_interface"]}\':'
                                               f'{forwarder["remote_port"]}' for forwarder in
                                               forwarders]) if forwarders else ''
            self.forwarders_ports = ', '.join([str(forwarder["local_port"])
                                               for forwarder in forwarders]) if forwarders else ''
            # ex. 587, 995
            self.all_host_ips = [gateway["host_ip"] for gateway in self.gateways]
            self.all_host_ips.append(self.destination["host_ip"])
            # ex. ['10.10.1.72', '10.10.2.92']
        except AttributeError as exp:
            self._logger.error('configuration file could not be parsed. %s', exp)  # raised by config.get()
            raise InvalidConfigurationFile
        except FileNotFoundError:  # open()
            self._logger.error('configuration file could not be opened ("FileNotFoundError")')
            raise InvalidConfigurationFile
        except IOError:  # open()
            self._logger.error('not enough permissions to open configuration file ("IOError")')
        except ValueError:  # json.loads()
            self._logger.error('JSON document could not be deserialized ("ValueError")')
            raise InvalidConfigurationFile
        except MultipleInvalid:  # Schema()
            self._logger.error('data structure (dict) validating failed ("MultipleInvalid")')
            raise InvalidConfigurationFile

    def get_config(self, filename):
        """Validates the data structure and parses the parameters in a dictionary."""
        with open(filename, 'r') as file:
            _temp = json.loads(file.read())
            _mode = MODE_SCHEMA(_temp)
            if _mode.get('mode') == 'TOR':
                schema = TOR_SCHEMA(_temp)
            elif _mode.get('mode') == 'FOR':
                schema = FOR_SCHEMA(_temp)
            elif _mode.get('mode') == 'PLAIN':
                schema = PLAIN_SCHEMA(_temp)
            else:
                self._logger.error('no mode enabled')
                schema = None
            return schema


def write_ssh_config_file(path_ssh_cfg_minitor, gateways, destination):
    """Writes the configuration file with ProxyJump directives for ssh.

    The IdentityFile cannot be given as a run-time parameter. Therefore,
    we resort to a directive in a config file

    Returns:
        bool: True on success, False otherwise.

    """
    # ex. [{'host_ip': '10.10.1.72', 'user': 'root', 'identity_file': '~/.ssh/id_rsa_be_vm'}]
    # ex. {'host_ip': '10.10.2.92', 'user':'root', 'identity_file': '~/.ssh/id_rsa_it_vm'}

    content = ''
    for gateway in gateways:
        content += f'Host {gateway["host_ip"]}\n' \
                   f'  HostName {gateway["host_ip"]}\n' \
                   f'  User {gateway["user"]} \n' \
                   f'  IdentitiesOnly yes \n' \
                   f'  IdentityFile {gateway["identity_file"]}\n\n'
    content += f'Host {destination["host_ip"]}\n' \
               f'  HostName {destination["host_ip"]}\n' \
               f'  User {destination["user"]} \n' \
               f'  IdentitiesOnly yes \n' \
               f'  IdentityFile {destination["identity_file"]}\n\n'
    LOGGER.debug('"%s" is written to ssh config config file: %s', content.replace("\n", ""),
                 path_ssh_cfg_minitor)
    try:
        with open(path_ssh_cfg_minitor, 'w') as data_source:
            data_source.write(content)
    except IOError:
        LOGGER.error('ssh config file %s could not be read', path_ssh_cfg_minitor)
        return False
    return True


class StateManager(LoggerMixin):  # context manager
    """Cleans up objects (eg. Tunnel, Assistant) when exiting.

    An KeyboardInterrupt, which is an exception, will be caught
    by this class, or specifically, by __exit__(). Consequently, this method
    will invoke the _clean_up() to stop all instantiated objects.
    """

    def __init__(self):
        super().__init__()
        self.running_instances = []  # this attribute was protected (_running...)

    def __enter__(self):
        self._logger.debug('Preparing to setup link...')
        return self

    def add_object(self, object_):
        """Collects instantiated Tunnel and Assistant, and shape(s) for clean up purposes."""
        self.running_instances.append(object_)

    def _clean_up(self):
        for object_ in reversed(self.running_instances):
            self._logger.debug('cleaning up %s', str(object_))
            object_.stop()
            sleep(0.05)  # sleep is introduced for the GUI package to have a cascade of disappearing canvas items

    def __exit__(self, exc_type, exc_value, exc_traceback):  # pylint: disable=inconsistent-return-statements
        self._logger.info('terminating Agent, Instructor, and then Tunnel...')
        self._clean_up()
        self._logger.info('cleaned up')
        if exc_type is KeyboardInterrupt:
            self._logger.debug("KeyboardInterrupt caught")
            return True  # https://effbot.org/zone/python-with-statement.htm


class Heartbeat(LoggerMixin):  # context manager
    """Determines periodically the state of the tunnel.

    An KeyboardInterrupt, which is an exception, will first be caught
    by this class. Consequently, this exception will stop the monitoring.
    """

    def __init__(self, local_heartbeat_port):
        super().__init__()
        self.thread = None
        self.is_tunnel_intact = True
        self.terminate = False
        self.local_heartbeat_port = local_heartbeat_port

    def _run_heartbeat(self):
        while self.thread.is_alive:
            if self.terminate:
                return None
            self.is_tunnel_intact = start_ping(self.local_heartbeat_port)
            if self.is_tunnel_intact:
                self._logger.debug('heartbeat signal was successfully returned')
            else:
                self._logger.error('heartbeat signal was not returned')
            sleep(HEARTBEAT_DURATION)

    def __enter__(self):
        self.thread = threading.Thread(target=self._run_heartbeat)
        self.thread.start()
        self._logger.info('heartbeat mechanism started')  # necessary to have log level "info"?
        return self

    def __exit__(self, type_, value, traceback):
        self.terminate = True
        self._logger.info('heartbeat mechanism stopped')  # necessary to have log level "info"?


def start_application(binary_name, binary_location):  # used in either FOR (w/ Thunderbird) or TOR mode (w/ Firefox)
    """Starts the application."""
    try:
        # process = subprocess.run([binary_location], capture_output=False)
        # process = subprocess.run([binary_location], subprocess.DEVNULL)
        # with open(os.devnull, 'w') as devnull:
        #     process = subprocess.run([binary_location], stdout=devnull, stderr=devnull)
        process = subprocess.Popen(binary_location, shell=True)
        # LOGGER.debug('application %s executed and the return code was: %s', binary_name, process.returncode)
        return process
    except FileNotFoundError:
        LOGGER.error('the executable binary %s of the application was not found', binary_name)
        return False


def start_ping(local_heartbeat_port):
    """Sends a HTTP GET request and processes the response."""
    heartbeat_port = local_heartbeat_port
    http_code = 0
    result = None
    try:
        with urllib.request.urlopen(f'http://localhost:{heartbeat_port}', timeout=2, data=None) as request_obj:
            http_code = request_obj.getcode()
    except (URLError, ConnectionResetError):
        LOGGER.debug('Agent did not respond to GET request. '
                     'probable cause: gateways or destination host unreachable, '
                     'localhost (client) has no connection to the Internet, or '
                     'Agent on destination host is not bind to local port.')
        result = False
    except timeout:  # this exception needed?
        LOGGER.debug('Agent did not respond to GET request in a timely fashion.'
                     'probable cause: gateways or destination host unreachable, '
                     'localhost (client) has no connection to the Internet, or '
                     'Agent on destination host is not bind to local port.')
        result = False
    if http_code == 200:
        result = True
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: transferagent.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for transferagent.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

NOTE: The TransferAgent class is responsible to purge the stream (ie. index in stream is at COMMAND_PROMPT)

"""

import inspect
import os.path
import pexpect
from .logging import LoggerMixin

__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''10-05-2019'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# Constant for Pexpect. This prompt is default for Fedora and CentOS.
COMMAND_PROMPT = '[#$] '


class TransferAgent(LoggerMixin):
    """Establishes a connection to the target destination host via one or more intermediaries."""

    def __init__(self, path_ssh_cfg, all_host_ips):
        """Initialize the TransferAgent object."""
        super().__init__()
        self.all_host_ips = all_host_ips
        self.last_host_ip = all_host_ips[-1]
        self.child = None
        self.path_ssh_cfg = path_ssh_cfg
        self.authenticated_hosts = []

    def __str__(self):
        return 'TransferAgent'

    @property
    def _path_to_agent_module(self):
        running_script = inspect.getframeinfo(inspect.currentframe()).filename
        running_script_dir = os.path.dirname(os.path.abspath(running_script))
        path_file = os.path.join(running_script_dir, 'payload', 'agent.py')
        self._logger.debug('agent.py resides in: %s', running_script_dir)
        return path_file

    def _generate_ssh_runtime_param(self):
        if len(self.all_host_ips) == 2:  # only 1 gateway + destination host
            # the result will be something in this format:
            # scp -F {} -o 'ProxyJump 10.10.1.72' /home/vincent/Pictures/andy_apollo_imdb.jpg 10.10.2.92:/tmp
            order_of_hosts = f'{self.all_host_ips[0]}'
        else:  # more than 1 gateway + destination host
            # the result will be something in this format:
            # scp -F {} -o 'ProxyJump 10.10.1.72,10.10.2.92' /home/vincent/Pictures/andy_apollo_imdb.jpg 10.10.3.82:/tmp
            order_of_hosts = ''
            for i, host in enumerate(self.all_host_ips):
                if i == 0:
                    order_of_hosts += f'{host}'
                elif i < len(self.all_host_ips) - 1:
                    order_of_hosts += f',{host}'
                else:
                    order_of_hosts += f' {host}'  # determine if this branch is still necessary.

        runtime_param = f"scp -v -F {self.path_ssh_cfg} -o 'ProxyJump {order_of_hosts}' " \
                        f"{self._path_to_agent_module} "
        runtime_param += f'{self.last_host_ip}:/tmp'
        self._logger.debug(runtime_param)
        return runtime_param

    def start(self):
        """Composes a SSH runtime param command and prints successful authentications."""
        result = False
        try:
            self.child = pexpect.spawn(self._generate_ssh_runtime_param(), env={"TERM": "dumb"}, timeout=10)
            # self.process.setecho(False)  # doesn't seem to have effect
            # self.process.waitnoecho()  # doesn't seem to have effect
            self._logger.debug('going through the stream to match patterns')
            for hostname in self.all_host_ips:
                # according to the documentation, "If you wish to read up to the end of the child's output -
                # without generating an EOF exception then use the expect(pexpect.EOF) method."
                # but apparently this doesn't work in a shell within a shell (SSH spawns a new shell)
                index = self.child.expect(
                    [f'Authenticated to {hostname}', 'Last failed login:', 'Last login:', 'socket error',
                     'not accessible', 'fingerprint', 'open failed: connect failed:', 'No such file', pexpect.TIMEOUT])
                if index == 0:
                    self._logger.info('authenticated to %s', hostname)  # logger level is "info" to inform user
                    self.authenticated_hosts.append(hostname)
                    result = True
                elif index == 1:
                    self._logger.debug('there were failed login attempts')
                    result = True
                elif index == 2:
                    self._logger.debug('there were no failed login attempts')
                    result = True
                elif index == 3:
                    self._logger.error('socket error. probable cause: SCP service on proxy or target machine disabled')
                    self.child.terminate()
                elif index == 4:
                    self._logger.error('the identity file is not accessible')
                    self.child.terminate()
                elif index == 5:
                    self._logger.warning('warning: hostname automatically added to list of known hosts')
                    self.child.sendline('yes')
                elif index == 6:
                    self._logger.error('SCP could not connect to %s', hostname)
                    self.child.terminate()
                elif index == 7:
                    pass
                elif index == 8:
                    self._logger.error('TIMEOUT exception was thrown. SCP could probably not connect to %s', hostname)
                    self.child.terminate()
                else:
                    self._logger.error('unknown state reached')
            self.child.expect(pexpect.EOF)  # the buffer has to be 'read' continuously, otherwise Pexpect deteriorates
        except pexpect.exceptions.ExceptionPexpect:
            self._logger.error('EOF is read; SCP has exited abnormally.')
            self.child.terminate()
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: tunnel.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for tunnel.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

NOTE: The Tunnel classes are responsible to purge the stream (ie. index in stream is at COMMAND_PROMPT)

"""

# from abc import ABC, abstractmethod
import threading
from time import sleep
import pexpect
from .logging import LoggerMixin

__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''10-05-2019'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# Constant for Pexpect. This prompt is default for Fedora and CentOS.
COMMAND_PROMPT = '[#$] '


class Tunnel(LoggerMixin):  # pylint: disable=too-many-instance-attributes
    """Establishes a connection to the target destination host via one or more intermediaries.

    Be aware, the child's buffer needs to be purged periodically. This can be done by invoking
    periodically_purge_buffer(). As verbose mode is enabled for SSH (the child process), it
    will slowly fill up the buffer, so this has to be taken care of. But don't invoke this
    method before having start()'ed BootstrapAgent.
    """

    def __init__(self, path_ssh_cfg, mode, all_host_ips, group_ports,  # pylint: disable=too-many-arguments
                 forward_connections=None, socks_server_address='localhost'):
        """Initialize the Tunnel object."""
        super().__init__()
        self.path_ssh_cfg = path_ssh_cfg
        self.mode = mode
        self.all_host_ips = all_host_ips
        self.last_host_ip = all_host_ips[-1]
        self.group_ports = group_ports
        self.forward_connections = forward_connections
        self.socks_server_address = socks_server_address
        self.child = None
        self.thread = None
        self.terminate = False
        self.authenticated_hosts = []

    def __str__(self):
        return 'Tunnel'

    def _generate_ssh_runtime_param(self):
        var_param = None
        # block below composes _mode specific_ forwarding strings
        if self.mode == 'FOR':
            var_param = f'{self.forward_connections} '
        elif self.mode == 'TOR':
            var_param = f'-L{self.group_ports["local_port_proxy"]}:{self.socks_server_address}:' \
                        f'{self.group_ports["remote_port_proxy"]} '
        elif self.mode == 'PLAIN':
            var_param = ''  # no *additional* ports will be forwarded

        if len(self.all_host_ips) == 2:
            # the result will be in this format: 'host1 host2' --> '10.10.1.72 10.10.2.92'
            order_of_hosts = f'{self.all_host_ips[0]} {self.all_host_ips[1]}'
        else:
            # the result will be something in this format: 'host1,host2 host3' --> '10.10.1.72,10.10.2.92 10.10.3.52'
            order_of_hosts = ''
            for i, host in enumerate(self.all_host_ips):
                if i == 0:
                    order_of_hosts += f'{host}'
                elif i < len(self.all_host_ips) - 1:
                    order_of_hosts += f',{host}'
                else:
                    order_of_hosts += f' {host}'  # is this branch necessary?

        # block below composes _base_ forwarding strings
        runtime_param = f'ssh -v -F {self.path_ssh_cfg} ' \
                        f'-L{self.group_ports["local_port_agent"]}:localhost:' \
                        f'{self.group_ports["remote_port_agent"]} ' \
                        f'-L{self.group_ports["local_port_heartbeat"]}:localhost:' \
                        f'{self.group_ports["remote_port_heartbeat"]} ' \
                        f'-L{self.group_ports["local_port_command"]}:localhost:' \
                        f'{self.group_ports["remote_port_command"]} ' \
                        f'-L{self.group_ports["local_port_transfer"]}:localhost:' \
                        f'{self.group_ports["remote_port_transfer"]} '
        runtime_param += var_param
        runtime_param += f'-J {order_of_hosts}'

        self._logger.debug(runtime_param)
        return runtime_param

    def start(self, debug=None):
        """Starts and controls SSH (child application) along with parameters.

        In addition, this method and mines for 'Authenticated' keywords, so
        we can keep track which hosts have been connected through.

        Args:
            debug(basestring): if True, TIMEOUT will not be raised and may block indefinitely. Use only for debugging
                                purposes to capture the output of the child, which is essentially, hidden 'under the
                                hood', and write it to a file.

        """
        result = False
        try:
            arguments = dict(command=self._generate_ssh_runtime_param(), env={"TERM": "dumb"}, encoding='utf-8')
            if debug:
                arguments.update(dict(timeout=10))
            self.child = pexpect.spawn(**arguments)
            # setecho() doesn't seem to have effect.
            #    doc says: Not supported on platforms where isatty() returns False.
            #    perhaps related to the recursive shells (SSH spawns a new shell in the current shell)
            self.child.setecho(False)
            self._logger.debug('going through the stream to match patterns')
            for hostname in self.all_host_ips:
                # according to the documentation, "If you wish to read up to the end of the child's output
                #    without generating an EOF exception then use the expect(pexpect.EOF) method."
                #    but apparently this doesn't work in a shell within a shell (SSH spawns a new shell)
                index = self.child.expect(
                    [f'Authenticated to {hostname}', 'Last failed login:', 'Last login:', 'socket error',
                     'not accessible', 'fingerprint', 'open failed: connect failed:', pexpect.TIMEOUT])
                if index == 0:
                    self._logger.info('authenticated to %s', hostname)  # logger level is "info" to inform user
                    self.authenticated_hosts.append(hostname)
                    result = True
                elif index == 1:
                    self._logger.debug('there were failed login attempts')
                    result = True
                elif index == 2:
                    self._logger.debug('there were no failed login attempts')
                    result = True
                elif index == 3:
                    self._logger.error('socket error. probable cause: SSH service on proxy or target machine disabled')
                    self.child.terminate()
                elif index == 4:
                    self._logger.error('the identity file is not accessible')
                    self.child.terminate()
                elif index == 5:
                    self._logger.warning('warning: hostname automatically added to list of known hosts')
                    self.child.sendline('yes')  # security issue
                elif index == 6:
                    self._logger.error('SSH could not connect to %s', hostname)
                    self.child.terminate()
                elif index == 7:
                    self._logger.error('TIMEOUT exception was thrown. SSH could probably not connect to %s', hostname)
                    self.child.terminate()
                else:
                    self._logger.error('unknown state reached')
            self.child.expect(COMMAND_PROMPT)
        except pexpect.exceptions.ExceptionPexpect:
            self._logger.error('EOF is read; SSH has exited abnormally')
            self.child.terminate()
        return result

    def stop(self):
        """Closes the SSH connection essentially by terminating the program SSH."""
        self.terminate = True
        if self.child.isalive():
            self._logger.debug('SSH is alive, terminating')
            self.child.terminate()
        self._logger.debug('SSH terminated')
        return True

    def debug(self):
        """Captures the output of the child (warning: BLOCKING)."""
        fout = open('~/mylog.txt', 'a')
        self.child.logfile = fout
        try:
            self.child.readlines()
        except pexpect.ExceptionPexpect:
            pass

    def periodically_purge_buffer(self):
        """Purges the child's (SSH) output buffer due to buffer limitations."""
        self.thread = threading.Thread(target=self._run_purger)
        self.thread.start()

    def _run_purger(self):
        while not self.terminate:
            try:
                self.child.expect([pexpect.TIMEOUT], timeout=0.2)
                sleep(2)
            except pexpect.exceptions.EOF:
                pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: agent.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for Agent.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import socket
import os
import logging.config
# import coloredlogs
import threading
import http.server
import socketserver
import json
import subprocess
import select
import sys
import atexit
import signal
# from dataclasses import dataclass
from abc import ABC
from struct import pack, unpack
from os.path import basename
from time import sleep

# Constants
PORT_AGENT = 44191  # the local port the agent uses to listen for instructions from Instructor.
# IMPORTANT NOTE: other ports are sent via an Instructor class which receives it from the CLI.

# Configuration for SOCKS proxy server classes
MAX_THREADS = 200
BUFSIZE = 2048
TIMEOUT_SOCKET = 5
VER = b'\x05'  # PROTOCOL VERSION 5
M_NOAUTH = b'\x00'  # '00' NO AUTHENTICATION REQUIRE
M_NOTAVAILABLE = b'\xff'  # 'FF' NO ACCEPTABLE METHODS
CMD_CONNECT = b'\x01'  # CONNECT '01'
ATYP_IPV4 = b'\x01'  # IP V4 address '01'
ATYP_DOMAINNAME = b'\x03'  # DOMAINNAME '03'

# Logging
LOGGER = logging.getLogger()  # not used?
LOGGER_BASENAME = '''Agent'''


class LoggerMixin:  # pylint: disable=too-few-public-methods
    """Contains a logger method for use by other classes."""

    def __init__(self):
        """Initialize the LoggerMixin object."""
        self._logger = logging.getLogger(f'{LOGGER_BASENAME}.{self.__class__.__name__}')


class TransferError(Exception):
    """Something went wrong during transfer of the file."""


def validate_transfer_server_response(response):
    """Validates the data structure of the content of an incoming HTTP request.

    These requests are received by the CommandServer and contains Linux commands.
    """
    if all([isinstance(response.get('process'), str),
            isinstance(response.get('status_code'), int)]):
        return response.get('process'), response.get('status_code')
    raise InvalidDataStructure


def validate_http_instruction(request):
    """Validates the data structure of the content of an incoming HTTP request.

    These requests are received by the Agent.
    """
    process = ['transfer_server_start', 'file_server_stop', 'proxy_server_start', 'proxy_server_stop',
               'heartbeat_responder_start', 'heartbeat_responder_stop', 'command_server_start', 'command_server_stop',
               'stop']  # 'authenticate_host'
    if not all([request.get('process') in process,
                isinstance(request.get('arguments'), dict)]):
        raise InvalidDataStructure
    return request.get('process'), request.get('arguments', {})


def validate_http_command(request):
    """Validates the data structure of the content of an incoming HTTP request.

    These requests are received by the CommandServer and contains Linux commands.
    """
    if isinstance(request.get('command'), str):
        return request
    raise InvalidDataStructure


class InvalidDataStructure(Exception):
    """The data structure is invalid."""

    def __init__(self):
        super().__init__("the data structure is invalid.")


class Agent(LoggerMixin):
    """Listens for instructions send by *Instructor."""

    def __init__(self, port):
        """Initialize the Agent object.

        Args:
            port (basestring): The local port used to listen for instructions from *Instructor.

        """
        super().__init__()
        self.port = port
        self.httpd = None
        self.terminate = False
        self.transfer_server = None
        self.command_server = None
        self.heartbeat_responder = None
        self.proxy_server = None
        self.authenticate = None

    def __str__(self):
        return 'Agent'

    def start(self):
        """Listens for incoming HTTP POST request."""
        self._logger.debug('starting agent')
        threading.Thread(target=self._watcher).start()
        agent = self

        class Handler(http.server.SimpleHTTPRequestHandler):
            """Parses HTTP requests."""

            logger_name = u'{base}.{suffix}'.format(base=LOGGER_BASENAME,
                                                    suffix='Handler')
            _logger = logging.getLogger(logger_name)
            socketserver.TCPServer.allow_reuse_address = True

            def do_POST(self):
                """Creates the response."""
                try:
                    data = self.rfile.read(int(self.headers['Content-Length']))  # b'{"process":"heartbeat_responder"}
                    self._logger.debug('the following request was received from *Instructor: %s', data)
                    instruction_string = data.decode('utf-8')  # convert byte to string (to original JSON document)
                    instruction_dict = json.loads(instruction_string)  # convert JSON to dict
                    # example: {'process': 'transfer_server_start', 'arguments': {'port': 44194}}
                    # validate structure of content of req:
                    process, arguments = validate_http_instruction(instruction_dict)
                except json.decoder.JSONDecodeError:  # json.loads()
                    self._logger.error('the content is incorrectly parsed in JSON')
                    process = 'default'  # to enter exit branch
                    arguments = {}
                except InvalidDataStructure:  # validate_http_instruction()
                    self._logger.error('data structure (dict) validation failed')
                    process = 'default'  # to enter exit branch
                    arguments = {}

                if getattr(agent, f'{process}')(**arguments):  # getattr(agent, transfer_server_start(port:44194))
                    self.send_response(200)
                    self.end_headers()
                    json_instruction = json.dumps({'result': True})
                    data = json_instruction.encode('utf-8')  # from string to byte
                    self.wfile.write(data)
                else:
                    self.send_response(200)
                    self.end_headers()
                    json_instruction = json.dumps({'result': False})
                    data = json_instruction.encode('utf-8')  # from string to byte
                    self.wfile.write(data)

        with socketserver.TCPServer(("localhost", self.port), Handler) as agent.httpd:
            self._logger.debug('serving at port %s', self.port)
            agent.httpd.serve_forever()

    def _watcher(self):
        while not self.terminate:
            sleep(1)
        self._logger.debug('sending shutdown()')
        self.httpd.shutdown()

    def heartbeat_responder_start(self, **kwargs):
        """Starts the heartbeat responder."""
        port = kwargs.get('port')
        self.heartbeat_responder = HeartbeatResponder(port)
        return self.heartbeat_responder.start()

    def transfer_server_start(self, **kwargs):
        """Starts the transfer server."""
        mode = kwargs.get('mode')
        port = kwargs.get('port')
        self.transfer_server = TransferServer(port)
        return self.transfer_server.start(mode)

    def command_server_start(self, **kwargs):
        """Starts the command server."""
        port = kwargs.get('port')
        self.command_server = CommandServer(port)
        return self.command_server.start()

    def proxy_server_start(self, **kwargs):
        """Starts the proxy server."""
        addr_i = kwargs.get('address_i')
        port_i = kwargs.get('port_i')
        addr_e = kwargs.get('address_e')
        self.proxy_server = ProxyServer(internal_address=addr_i,
                                        internal_port=port_i,
                                        external_address=addr_e)
        return self.proxy_server.start()

    def heartbeat_responder_stop(self):
        """Stops the heartbeat responder."""
        return True if self.heartbeat_responder is None else self.heartbeat_responder.stop()

    def transfer_server_stop(self):
        """Stops the transfer server."""
        return True if self.transfer_server is None else self.transfer_server.stop()

    def command_server_stop(self):
        """Stops the command server."""
        return True if self.command_server is None else self.command_server.stop()

    def proxy_server_stop(self):
        """Stops the proxy server."""
        return True if self.proxy_server is None else self.proxy_server.stop()

    def stop(self):
        """Stops the Agent."""
        self._logger.debug('stopping heartbeat responder, and transfer, command and proxy server, if running...')
        self.heartbeat_responder_stop()
        self.proxy_server_stop()
        self.command_server_stop()
        self.transfer_server_stop()
        self._logger.debug('finally, stopping agent...')
        self.terminate = True
        return True

    def default(self):
        """Stops the Agent."""
        self._logger.debug('the process is unknown')
        # self.stop()


class TransferServer(LoggerMixin):
    """Receives file(s) send by *Instructor()."""

    def __init__(self, port):
        """Initializes the TransferServer object."""
        super().__init__()
        self.port = port
        self.data_protocol = None
        self.terminate = False

    def start(self, mode, src_file_path=None, dest_path=None):
        """Starts transfer server."""
        self._logger.debug('starting transfer server')
        self.data_protocol = DataProtocol(mode, self.port)
        self.data_protocol.start()
        if mode == 'receive':
            threading.Thread(target=self._receive_file).start()
        elif mode == 'send':
            return False  # not implemented, yet.
        return True

    def stop(self):
        """Stops transfer server."""
        self._logger.debug('stopping transfer server...')
        self.terminate = True
        self.data_protocol.stop()
        return True

    def _receive_file(self):
        # result = False
        try:
            while not self.terminate:
                metadata = self.data_protocol.receive_metadata()
                self.data_protocol.receive_file(metadata)
                sleep(1)  # Costas, is this necessary to elevate performance?
        except TransferError:
            self._logger.error('?')

    def _send_file(self, src_file_path, dest_path):
        result = False
        try:
            result = all([self.data_protocol.send_metadata(src_file_path, dest_path),
                          self.data_protocol.send_file(src_file_path)])
            result = True
        except FileNotFoundError:
            self._logger.error('file or directory does not exist')
        except TransferError:
            self._logger.error('something went wrong during transfer')
        finally:
            return result


class SocketServer(LoggerMixin):
    """Manages sockets."""

    def __init__(self):
        super().__init__()
        self.terminate = False
        self.socket_ = None
        self.connections = []

    def create_socket(self):
        """Create a new INET (IPv4), STREAMing socket (TCP)."""
        try:
            self.socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as err:
            self._logger.debug("failed to create socket: %s", err)

    def create_socket_and_listen(self, port, host='localhost'):
        """Create a new INET, STREAMing socket and listens for connections made to the socket."""
        try:
            self.socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as err:
            self._logger.debug("failed to create socket: %s", err)
        # sockets are interior endpoints built for sending and receiving data.
        self.socket_.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # to avoid "Address already in use" error, set the SO_REUSEADDR flag.
        # this tells the kernel o reuse a local socket in TIME_WAIT state,
        # without waiting for its natural timeout to expire.
        self.socket_.bind((host, port))
        # socket.bind(address) binds the socket to address.
        self._logger.debug('serving at port %s', port)
        self.socket_.listen(2)
        # listen(backlog) specifies the number of unaccepted connections that the -
        # system will allow before refusing new connections.
        self._logger.debug('waiting for a connection...')

    def handle_connections(self):
        """Accepts connections and add them to an instance var of type called connections."""

        def _conn_handler():
            while not self.terminate:  # keep listening to new connections
                try:
                    connection, _ = self.socket_.accept()  # blocking. calling close() will raise exception
                    self._logger.debug('connection received from %s to %s',
                                       connection.getsockname(), connection.getpeername())
                    self.connections.append(connection)
                except OSError:
                    pass

        threading.Thread(target=_conn_handler).start()

    def close_socket(self):  # https://stackoverflow.com/questions/409783/socket-shutdown-vs-socket-close
        """Closes socket."""
        self.terminate = True
        self.socket_.shutdown(socket.SHUT_RDWR)
        # closes the underlying connection and sends a FIN / EOF to the peer regardless of how many processes have
        # handles to the socket. However, it does not deallocate the socket and you still need to call close afterwards
        self.socket_.close()
        # it decrements the handle count by one and if the handle count has reached zero then the socket and associated
        # connection goes through the normal close procedure (effectively sending a FIN / EOF to the peer) and the
        # socket is deallocated. The thing to pay attention to here is that if the handle count does not reach zero
        # because another process still has a handle to the socket then the connection is not closed and the socket is
        # not deallocated.

    def connect_to_socket(self, address, port, external_address=None):
        """Connect to a remote socket at given address."""
        if external_address:
            try:
                self.socket_.setsockopt(  # what does this actually do?
                    socket.SOL_SOCKET,
                    socket.AF_INET,
                    external_address.encode()
                )
            except socket.error as exc:
                self._logger.info("something went wrong when setting socket operation: %s ", exc)
        try:
            self.socket_.connect((address, port))  # tuple
            self._logger.debug("connection established from %s to %s",
                               self.socket_.getsockname(),
                               self.socket_.getpeername())
            return True
        except socket.error as exc:
            self._logger.info("connection could not be established to %s ", exc)
            return False


class DataProtocol(SocketServer):  # Costas, LoggerMixin is subclassed by SocketServer, but I'd like to mention it here
    """Dictates how to format, transmit and receive data.

    Encodes file metadata and sends it along with the content of the (binary) file
    or decodes file metadata and writes received data to a new file.
    """

    def __init__(self, mode, port):
        """Initializes the DataProtocol object."""
        super().__init__()
        self.mode = mode
        self.port = port

    def start(self):
        """Starts the data protocol in either receiving of sending mode."""
        if self.mode == 'receive':
            self.create_socket_and_listen(self.port)
            self.handle_connections()
        elif self.mode == 'send':
            self.create_socket()
            self.connect_to_socket('localhost', self.port)

    def stop(self):
        """Closes the socket."""
        self.close_socket()

    def send_metadata(self, source_file_path, destination_path, padding=16):
        """Encodes the metadata."""
        metadata = {'dest_path': destination_path,
                    'file_name': basename(source_file_path),
                    'file_size': str(os.path.getsize(source_file_path))
                    }
        self.socket_.sendall(bytes('metadata', 'utf-8'))  # string is 8 bytes
        for value in metadata.values():
            self._logger.debug('convert metadata of %s to bytes and send', value)
            length = bin(len(value))[2:].zfill(padding)  # from decimal to binary (eg. 0000000000001110)
            data = bytes(length, 'utf-8')  # turn into byte (eg. b'00000000001110')
            data += bytes(value, 'utf-8')  # turn into byte (eg. b'amsterdam.jpg')
            self.socket_.sendall(data)
        return self._check_delivery_code()

    def send_file(self, source_file_path):
        """Sends the content of the file."""
        self._logger.debug('convert file %s to bytes and send', basename(source_file_path))
        self.socket_.sendall(bytes('filedata', 'utf-8'))
        data = open(source_file_path, 'rb')  # type is "_io.BufferedReader"
        self.socket_.sendall(data.read())
        return self._check_delivery_code()

    def receive_metadata(self):
        """Decodes the metadata."""
        if not self.connections:
            return
        connection = self.connections[-1]
        process = connection.recv(8)
        if process != b'metadata':
            return
        metadata = {'dest_path': None,
                    'file_name': None,
                    'file_size': None
                    }
        for key, value in metadata.items():
            try:
                length_binary = connection.recv(16)
                length_int = int(length_binary, 2)
                metadata[key] = connection.recv(length_int)
                self._logger.debug('%s: %s received' % (key, metadata[key]))
            except IOError:
                return self._send_status_code(connection, 'metadata', 1)
        self._send_status_code(connection, 'metadata', 0)
        return metadata  # {'dest_path': b'/tmp', 'file_name': b'amsterdam.jpg', 'file_size': b'98130'

    def receive_file(self, metadata):
        """Writes the received data to a file."""
        if not metadata:
            return
        connection = self.connections.pop()
        process = connection.recv(8)
        if process != b'filedata':
            return
        file_size = int(metadata['file_size'])
        try:
            file_to_write = open(os.path.join(metadata['dest_path'],
                                              metadata['file_name']), 'wb')  # can raise: "No such file or directory"
            chunk_size = 4096
            while file_size > 0:
                if file_size < chunk_size:
                    chunk_size = file_size
                    self._logger.debug('receiving last chunk of data...')
                data = connection.recv(chunk_size)
                file_to_write.write(data)
                file_size -= len(data)
            file_to_write.close()
            self._logger.debug('received all chunks of data')
        except OSError:  # raised by open() or socket()
            return self._send_status_code(connection, 'filedata', 1)
        return self._send_status_code(connection, 'filedata', 0)

    def _check_delivery_code(self):
        status_code = self.socket_.recv(41)
        status_string = status_code.decode("utf-8")  # from bytes to string (containing JSON doc)
        status_string_dict = json.loads(status_string)  # from JSON doc to dict
        process, status_code = validate_transfer_server_response(status_string_dict)
        self._logger.debug('status code from transfer server: %s: %s', process, status_code)
        result = False
        if process == 'metadata' and status_code == 0:
            self._logger.debug('metadata is transferred')
            result = True
        elif process == 'filedata' and status_code == 0:
            self._logger.debug('file is transferred')
            result = True
        return result

    def _send_status_code(self, connection, process, status_code):
        if process == 'metadata' and status_code == 0:
            self._logger.debug('metadata is received successfully')
        elif process == 'filedata' and status_code == 0:
            self._logger.debug('filedata is received successfully')
        else:
            self._logger.debug('something went wrong')
        msg = {'process': f'{process}', 'status_code': int(status_code)}
        json_instruction = json.dumps(msg)  # from dict to JSON
        data = json_instruction.encode('utf-8')  # from string to byte
        try:
            connection.sendall(data)
            return True
        except OSError:
            return False


class CommandServer(LoggerMixin):
    """Listens for Linux commands send by *Instructor() and responds with result."""

    #  determine first whether port is bind or not
    #  < code >
    #  self._logger.error('Port already bind. Probably by having executed this method twice.')

    def __init__(self, port):
        """Initialize the CommandServer object.

        Args:
            port (basestring): <>

        """
        super().__init__()
        self.port = port
        self.httpd = None
        self.terminate = False

    def start(self):
        """Listens for connections."""
        self._logger.debug('starting command server...')
        threading.Thread(target=self._watcher).start()
        threading.Thread(target=self._serve).start()
        return True

    def _serve(self):
        """Listens for incoming HTTP POST request."""
        instance = self

        class Handler(http.server.SimpleHTTPRequestHandler):
            """Parses HTTP requests."""

            logger_name = u'{base}.{suffix}'.format(base=LOGGER_BASENAME,
                                                    suffix='Handler')
            _logger = logging.getLogger(logger_name)
            # socketserver.BaseRequestHandler.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            socketserver.TCPServer.allow_reuse_address = True

            def do_POST(self):  # pylint: disable=invalid-name
                """Creates the response containing the result."""
                try:
                    data = self.rfile.read(int(self.headers['Content-Length']))  # b'{"command":"ls -l /"}
                    self._logger.debug('following raw Linux command was received from instructor: %s', data)
                    command_string = data.decode('utf-8')  # convert byte to string in JSON format
                    command_dict = json.loads(command_string)  # convert JSON to dict
                    command_val = validate_http_command(command_dict)  # validate the structure of the content of req.
                    command = command_val.get('command')
                    instance._logger.debug('following Linux command was received from instructor: %s', command)
                    # eg. b'{"command": "hostname"}'
                except json.decoder.JSONDecodeError:  # json.loads()
                    self._logger.error('the content is incorrectly parsed in JSON')
                    return False
                except InvalidDataStructure:  # validate_http_instruction()
                    self._logger.error('data structure (dict) validation failed')
                    return False
                result_command = instance._issue_command(command.split())
                self.send_response(200)
                self.end_headers()
                self.wfile.write(result_command)
                result = True
                return result

        with socketserver.TCPServer(("localhost", self.port), Handler) as instance.httpd:
            self._logger.debug('serving at port %s', self.port)
            instance.httpd.serve_forever()

    def _watcher(self):
        while True:
            if self.terminate:
                self._logger.debug('sending shutdown()')
                self.httpd.shutdown()
                break

    def _issue_command(self, command):
        result = b'ERROR: command not recognized'
        try:
            result = subprocess.check_output(command).rstrip()
            self._logger.debug('result of Linux commando is %s', result)  # eg. b'server.enterprise.com'
        except FileNotFoundError:
            self._logger.error('Linux command could not be executed')
        finally:
            return result

    def stop(self):
        """Stops the command server."""
        self._logger.debug('stopping command server...')
        self.terminate = True
        return True


class HeartbeatResponder(LoggerMixin):
    """Responds to GET requests from powermolecli with HTTP code 200."""

    #  determine first whether port is bind or not
    #  self._logger.error('Port already bind. Probably by having executed this method twice.')

    def __init__(self, port):
        super().__init__()
        self.port = port
        self.httpd = None
        self.terminate = False

    def start(self):
        """Executes the HTTP server that responds to GET requests (heartbeats) from powermolecli."""
        self._logger.debug('starting heartbeat responder...')
        threading.Thread(target=self._watcher).start()
        threading.Thread(target=self._serve).start()
        return True

    def _serve(self):
        instance = self

        class Handler(http.server.SimpleHTTPRequestHandler):
            """Parses HTTP requests."""

            def __init__(self, *args, **kwargs):
                # kwargs['directory'] = directory  # I had to comment out, because -->
                super().__init__(*args, **kwargs)  # --> "TypeError: __init__ got an unexpected kwarg 'directory'"

            def do_GET(self):
                """Creates the response."""
                instance._logger.debug('GET request received')
                self.send_response(200)
                self.end_headers()

        with socketserver.TCPServer(("localhost", self.port), Handler) as instance.httpd:
            self._logger.debug('serving at port %s', self.port)
            instance.httpd.serve_forever()

    def _watcher(self):
        while True:
            if self.terminate:
                self._logger.debug('sending shutdown()')
                self.httpd.shutdown()
                break

    def stop(self):
        """Terminates the HTTP server responsible for responding to GET request (heartbeats) from powermolecli."""
        self._logger.debug('stopping heartbeat responder...')
        self.terminate = True
        return True


# Daemon in Python
# from Costas Tyfoxylos costas.tyf@gmail.com
# 2019-10-17: Refactored by myself (Vincent Schouten)

# Sources:
# [1] Based on http://web.archive.org/
#                        web/20131025230048/http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
#   The changes are:
#   1 - Uses file open context managers instead of calls to file().
#   2 - Forces stdin to /dev/null. stdout and stderr go to log files.
#   3 - Uses print instead of sys.stdout.write prior to pointing stdout to the log file.
#   4 - Omits try/excepts if they only wrap one error message w/ another.
# [2] https://stackoverflow.com/questions/33560802/pythonhow-os-fork-works
# [3] https://stackoverflow.com/questions/8777602/why-must-detach-from-tty-when-writing-a-linux-daemon


class Daemon(ABC):
    """Instantiates the daemon."""

    def __init__(self, pid_file=None, stdout=None, stderr=None):
        self.stdout = stdout or './daemon_out.log'
        self.stderr = stderr or './daemon_err.log'
        self.pid_file = pid_file or './daemon.pid'

    def _remove_pid(self):
        """Deletes the pid file."""
        os.remove(self.pid_file)

    def _daemonize(self):
        """Double forking of the process."""
        # fork 1 to spin off the child that will spawn the daemon.
        if os.fork() > 0:
            sys.exit(0)  # exit first parent
        # This is the child.

        # 1. clear the session id to clear the controlling TTY.
        # 2. set the umask so we have access to all files created by the daemon.
        os.setsid()
        os.umask(0)

        # fork 2 ensures we can't get a controlling TTY [ttd]?
        if os.fork() > 0:
            sys.exit(0)  # exit from second parent
        # This is a child that can't ever have a controlling TTY.

        # redirect standard file descriptor for *stdin* (essentially shut down stdin)
        with open('/dev/null', 'r') as dev_null:
            os.dup2(dev_null.fileno(), sys.stdin.fileno())  # os.dup <-- duplicate file descriptor

        # redirect standard file descriptor for *stderr* to log file
        sys.stderr.flush()
        with open(self.stderr, 'a+') as stderr:
            os.dup2(stderr.fileno(), sys.stderr.fileno())  # os.dup <-- duplicate file descriptor

        # redirect standard file descriptor for *stdout* to log file
        sys.stdout.flush()
        with open(self.stdout, 'a+') as stdout:
            os.dup2(stdout.fileno(), sys.stdout.fileno())  # os.dup <-- duplicate file descriptor

        # registered functions are executed automatically when the interpreter session is terminated normally.
        atexit.register(self._remove_pid)

        #   py interpreter
        #    |
        #   (fork) < duplicate itself
        #    |
        #    ├─ parent < exit this process!
        #    |
        #   (setsid) < detach from the terminal (ie. no controlling TTY) to avoid certain signals
        #    |
        #   (fork) < duplicate itself
        #    |
        #    ├─ parent < exit this process!
        #    |
        #    └─ child < store the pid of this process
        #
        pid = str(os.getpid())

        # write pid to file
        with open(self.pid_file, 'w') as pid_f:
            pid_f.write('{0}'.format(pid))

    @property
    def pid(self):
        """Returns the pid read from the pid file."""
        try:
            with open(self.pid_file, 'r') as pid_file:
                pid = int(pid_file.read().strip())
            return pid
        except IOError:
            return

    def start(self, function):
        """Starts the daemon."""
        # print('Starting...')
        if self.pid:
            print(('PID file {0} exists. '
                   'Is the daemon already running?').format(self.pid_file))
            sys.exit(1)
        self._daemonize()
        function()

    def stop(self):
        """Stops the daemon."""
        if not self.pid:
            print(("PID file {0} doesn't exist. "
                   "Is the daemon not running?").format(self.pid_file))
            return
        try:
            while 1:
                os.kill(self.pid, signal.SIGTERM)
                sleep(1)
        except OSError as err:
            if 'No such process' in err.strerror and \
                    os.path.exists(self.pid_file):
                os.remove(self.pid_file)
            else:
                print(err)
                sys.exit(1)

    def restart(self, function):
        """Restarts the daemon."""
        self.stop()
        self.start(function)


# Small Socks5 Proxy Server in Python
# from https://github.com/MisterDaneel/
# 2019-10-17: Refactored as OOP by Vincent Schouten (developer of powermole)
# 2021-05-16: Heavily refactored by Vincent Schouten


class ProxyServer(LoggerMixin):
    """De-encapsulating incoming SOCKS connections from client and forward data to destination host.

    SOCKS is a generic proxy protocol for TCP/IP-based networking applications.

    Authentication method = *no* authentication - username/password is not supported!
    """

    def __init__(self, internal_address, internal_port, external_address=None):
        """Initializes a Proxy object."""
        super().__init__()
        self.internal_address = internal_address
        self.internal_port = internal_port
        self.external_address = external_address
        self.socket_server = SocketServer()  # Costas, composition OK?
        self.should_terminate = False

    def start(self):
        """Starts the SOCKS proxy server."""
        self._logger.debug('starting SOCKS proxy server...')
        self.socket_server.create_socket_and_listen(self.internal_port, self.internal_address)
        self.socket_server.handle_connections()
        threading.Thread(target=self._thread_main_method_per_conn).start()
        return True

    def stop(self):
        """Stops the SOCKS proxy server."""
        self._logger.info("stopping SOCKS proxy server...")
        self.should_terminate = True
        self.socket_server.close_socket()  # this will (also) stop the threaded handle_connections() in SocketServer()
        return True

    def _thread_main_method_per_conn(self):
        while not self.should_terminate:
            if threading.activeCount() > MAX_THREADS:
                sleep(2)
                continue
            if not self.socket_server.connections:
                sleep(1)
                continue
            connection = self.socket_server.connections.pop(0)
            threading.Thread(target=self._negotiate_and_forward_data, args=(connection,)).start()

    def _negotiate_and_forward_data(self, connection):
        """Identifies SOCKS request and forwards data from client to destination.

        o Subnegotiation() is responsible for negotiating version (5) and method (NO AUTH).
        o Communicate() is responsible for forwarding data to destination which can be found in request message.

        """
        self._logger.debug("start SOCKS subnegotiation")
        if Subnegotiation(connection).start():
            self._logger.debug("start handling SOCKS request and forward data")
            Communicate(connection, self.external_address, self.should_terminate).start()


class Communicate(LoggerMixin):
    """Handles SOCKS requests from client, sends replies and forwards data."""

    def __init__(self,
                 connection,
                 external_address,
                 should_terminate):
        """Initializes an Communicate object."""
        super().__init__()
        self.connection = connection
        self.external_address = external_address
        self.should_terminate = should_terminate
        self.socket_server = SocketServer()

    def _proxy_loop(self):
        """Forwards data from client to destination.

        https://steelkiwi.com/blog/working-tcp-sockets/
        Here we call select.select to ask the OS to check given sockets
        whether they are ready to write, read, or if there is some exception respectively.
        That is why it passes three lists of sockets to specify which socket is expected
        to be writable, readable, and which should be checked for errors. This call will
        block the program (unless a timeout argument is passed) until some of the passed sockets are ready

        The select function blocks the thread until data is available
        on a specified socket. Then the data is forwarded to the right recipient.
        """
        self._logger.debug("data from client is being forwarded to destination")
        while not self.should_terminate:
            try:
                reader, _, _ = select.select([self.connection, self.socket_server.socket_], [], [], 1)
            except select.error as err:
                self._logger.debug('select failed: %s', err)
                return
            if not reader:
                continue
            try:
                for sock in reader:
                    data = sock.recv(BUFSIZE)
                    if not data:
                        return
                    if sock is self.socket_server.socket_:
                        self.connection.send(data)
                    else:
                        self.socket_server.socket_.send(data)
            except socket.error as err:
                self._logger.debug('proxy loop failed: %s', err)
                return

    def _get_destination_address(self):
        """Returns the destination address and port found in the SOCKS request.

        The SOCKS request information is sent by the client as soon as it has
        established a connection to the SOCKS server and completed the
        method-dependent subnegotiation.

        The SOCKS request is formed as follows:
        +----+-----+-------+------+----------+----------+
        |VER | CMD |  RSV  | ATYP | DST.ADDR | DST.PORT |
        +----+-----+-------+------+----------+----------+
        | 1  |  1  | X'00' |  1   | Variable |    2     |
        +----+-----+-------+------+----------+----------+

        Where:
        o  VER    protocol version: X'05'
        o  CMD
          o  CONNECT X'01'
          o  BIND X'02'
          o  UDP ASSOCIATE X'03'
        o  RSV    RESERVED
        o  ATYP   address type of following address
          o  IP V4 address: X'01'
          o  DOMAINNAME: X'03'
          o  IP V6 address: X'04'
        o  DST.ADDR     desired destination address
        o  DST.PORT     desired destination port in network octet order

        """
        try:
            s5_request = self.connection.recv(BUFSIZE)
        except ConnectionResetError as exp:
            if self.connection != 0:
                self.connection.close()
            self._logger.debug("something went wrong: %s", exp)
            return None
        if (
            s5_request[0:1] != VER or
            s5_request[1:2] != CMD_CONNECT or
            s5_request[2:3] != b'\x00'
        ):
            return None
        # IPV4
        if s5_request[3:4] == ATYP_IPV4:
            dst_addr = socket.inet_ntoa(s5_request[4:-2])
            dst_port = unpack('>H', s5_request[8:len(s5_request)])[0]
        # DOMAIN NAME
        elif s5_request[3:4] == ATYP_DOMAINNAME:
            sz_domain_name = s5_request[4]
            dst_addr = s5_request[5: 5 + sz_domain_name - len(s5_request)]
            port_to_unpack = s5_request[5 + sz_domain_name:len(s5_request)]  # b''
            dst_port = unpack('>H', port_to_unpack)[0]  # b''
        else:
            return None
        return dst_addr, dst_port

    def _send_reply(self, destination_address):
        """Sends a reply in response to a request.

        In the reply to a CONNECT, BND.PORT contains the port number that the
        server assigned to connect to the target host, while BND.ADDR
        contains the associated IP address.

        Server reply:
        +----+-----+-------+------+----------+----------+
        |VER | REP |  RSV  | ATYP | BND.ADDR | BND.PORT |
        +----+-----+-------+------+----------+----------+

        Where:
            o  REP    Reply field:
                o  X'00' succeeded
                o  X'01' general SOCKS server failure
                o  X'02' connection not allowed by ruleset
                o  X'03' Network unreachable
                o  X'04' Host unreachable
                o  X'05' Connection refused
                o  X'06' TTL expired
                o  X'07' Command not supported
                o  X'08' Address type not supported
                o  X'09' to X'FF' unassigned
            o  ATYP   address type of following address
                o  IP V4 address: X'01'  # length 4 octets
                o  DOMAINNAME: X'03'  # first octet of the field contains the number of octets of name that follow
                o  IP V6 address: X'04'  # length 16 octets

        """
        result = False
        rep = b'\x07'
        bnd = b'\x00' + b'\x00' + b'\x00' + b'\x00' + b'\x00' + b'\x00'
        if destination_address:
            address, port = destination_address
            self.socket_server.create_socket()
            status = self.socket_server.connect_to_socket(address, port, self.external_address)
            if not status:
                self.socket_server.socket_ = 0
        if not destination_address or self.socket_server.socket_ == 0:
            rep = b'\x01'  # failure
        else:
            rep = b'\x00'  # success
            bnd = socket.inet_aton(self.socket_server.socket_.getsockname()[0])
            bnd += pack(">H", self.socket_server.socket_.getsockname()[1])
        reply = VER + rep + b'\x00' + ATYP_IPV4 + bnd
        try:
            self.connection.sendall(reply)
            if rep == b'\x00':
                result = True
        except socket.error:
            if self.connection != 0:
                self.connection.close()
            result = False
        finally:
            return result

    def start(self):
        """Starts communicating."""
        destination_address = self._get_destination_address()
        if self._send_reply(destination_address):
            self._proxy_loop()
        else:
            if self.connection != 0:
                self.connection.close()


class Subnegotiation(LoggerMixin):
    """Negotiates version identifier (5) and method selection (NO AUTH) with client."""

    def __init__(self, connection):
        """Initializes a Subnegotiation object."""
        super().__init__()
        self.connection = connection

    def _retrieve_ver_and_method(self):
        """Retrieves version identifier/method selection.

        The client connects to the SOCKS proxy server, and sends
        a version identifier/method selection message

        Client version identifier/method selection message:
            +----+----------+----------+
            |VER | NMETHODS | METHODS  |
            +----+----------+----------+
            | 1  |    1     | 1 to 255 |
            +----+----------+----------+

        VER field is set to X'05'

        The values currently defined for METHOD(S) are:
          o  X'00' NO AUTHENTICATION REQUIRED
          o  X'01' GSSAPI
          o  X'02' USERNAME/PASSWORD
          o  X'03' to X'7F' IANA ASSIGNED
          o  X'80' to X'FE' RESERVED FOR PRIVATE METHODS
          o  X'FF' NO ACCEPTABLE METHODS

        """
        try:
            method_select_packet = self.connection.recv(BUFSIZE)
        except socket.error:
            self._logger.debug("Error")
            return M_NOTAVAILABLE  # b'\xff'
        if VER != method_select_packet[0:1]:
            return M_NOTAVAILABLE
        nmethods = method_select_packet[1]  # The NMETHODS field contains the number of method identifier octets -
        methods = method_select_packet[2:]  # that appear in the METHODS field.
        if len(methods) != nmethods:
            return M_NOTAVAILABLE
        for method in methods:
            if method == ord(M_NOAUTH):
                return M_NOAUTH  # X'00' NO AUTHENTICATION REQUIRED
        return M_NOTAVAILABLE

    def _send_sel_method(self, method):
        """Sends selected method.

        The server selects from one of the methods given in METHODS, and
        sends a METHOD selection message.

        Server method selection message:
        +-----+--------+
        | VER | METHOD |
        +-----+--------+
        |  1  |   1    |
        +----+---------+

        """
        if method != M_NOAUTH:
            return False
        reply = VER + method
        try:
            self.connection.sendall(reply)
        except socket.error as exp:
            self._logger.debug("something unexpected happened: %s", exp)
            return False
        self._logger.debug("method-dependent subnegotiation has completed")
        return True

    def start(self):
        """Starts subnegotiation."""
        method = self._retrieve_ver_and_method()
        return self._send_sel_method(method)


def main():
    """Main method."""
    logging.basicConfig(level='DEBUG',
                        filename='/tmp/log',
                        filemode='w',
                        format='%(asctime)s %(name)s %(levelname)s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')
    # coloredlogs.install(level='DEBUG')
    agent = Agent(PORT_AGENT)
    try:
        agent.start()
    except KeyboardInterrupt:
        agent.stop()
        raise SystemExit(0)


if __name__ == "__main__":
    if len(sys.argv) == 2:
        deploy_path = sys.argv[1]
        pid_file = os.path.join(deploy_path, 'agent.pid')
        stdout = os.path.join(deploy_path, 'daemon_out.log')
        stderr = os.path.join(deploy_path, 'daemon_err.log')
        d = Daemon(pid_file=pid_file, stdout=stdout, stderr=stderr)
        d.start(main)
        # main()
    else:
        print("no working path given")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: instructor.py
#
# Copyright 2021 Vincent Schouten
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for instructor.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

# import socket
# import os.path
import urllib.request
import json
# import threading
from socket import timeout
from abc import ABC, abstractmethod
# from os.path import basename
from urllib.error import URLError
from voluptuous import Schema, Required, Any, MultipleInvalid
from .logging import LoggerMixin
from .payload.agent import DataProtocol
from .powermolelibexceptions import InvalidDataStructure

__author__ = '''Vincent Schouten <inquiry@intoreflection.co>'''
__docformat__ = '''google'''
__date__ = '''10-05-2019'''
__copyright__ = '''Copyright 2021, Vincent Schouten'''
__credits__ = ["Vincent Schouten"]
__license__ = '''MIT'''
__maintainer__ = '''Vincent Schouten'''
__email__ = '''<inquiry@intoreflection.co>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# Constant for Pexpect. This prompt is default for Fedora and CentOS.
COMMAND_PROMPT = '[#$] '

# payload
HTTP_RESPONSE = Schema({Required("result"): Any(True, False)})


def validate_transfer_server_response(response):
    """Validates the data structure of the content of an incoming HTTP request.

    These requests are received by the CommandServer and contains Linux commands.
    """
    if all([isinstance(response.get('process'), str),
            isinstance(response.get('status_code'), int)]):
        return response.get('process'), response.get('status_code')
    raise InvalidDataStructure


class Instructor(ABC, LoggerMixin):
    """Models an Instructor to interact with the Agent residing on target destination host.

    Note: As the Agent sits on top of target destination hosts' OS, many functions can be
    performed more effectively.

    """

    def __init__(self, group_ports):
        """Initializes the Instructor object."""
        super().__init__()
        self.group_ports = group_ports
        self.host = '127.0.0.1'  # if-card of client
        self.socket_ = None

    def __str__(self):
        return 'Instructor'

    def _send_instruction(self, instruction):
        json_instruction = json.dumps(instruction)  # serialize dict to a JSON formatted str
        data = json_instruction.encode('utf-8')  # encode JSON formatted string to byte
        try:
            with urllib.request.urlopen(f'http://{self.host}:{self.group_ports["local_port_agent"]}',
                                        timeout=5,
                                        data=data) as request_obj:
                response_string = request_obj.read().decode('utf-8')  # from byte to string (JSON format)
            response_dict = json.loads(response_string)  # from JSON to dict
            response = HTTP_RESPONSE(response_dict)  # validating the structure of the content of an HTTP request
            result = response.get('result')
        except URLError:  # urllib.request.urlopen()
            self._logger.error('something went wrong when instructing the Agent, probably host unreachable')
            result = False
        except ConnectionResetError:  # urllib.request.urlopen()
            self._logger.error('connection is reset: Agent is probably not alive')
            result = False
        except timeout:  # urllib.request.urlopen()
            self._logger.error('timeout exceeded when instructing the Agent')
            result = False
        except json.decoder.JSONDecodeError:  # json.loads()
            self._logger.error('response of Agent could not be read, '
                               'JSON document could not be deserialized')
            result = False
        except MultipleInvalid:  # HTTP_RESPONSE()
            self._logger.error('response of Agent could not be read. '
                               'data structure validating failed ("MultipleInvalid")')
            result = False
        return result

    def _start_heartbeat_responder(self, host_port):
        """Sends the instruction to Agent to start the heartbeat responder."""
        return self._start_server(host_port, 'heartbeat_responder_start', 'heartbeat responder')

    def _start_command_server(self, host_port):
        """Starts the command server."""
        return self._start_server(host_port, 'command_server_start', 'command server')

    def _start_transfer_server(self, host_port):  # Costas, I've changed this method
        """Starts the transfer server."""
        # return self._start_server(host_port, 'transfer_server_start', 'transfer server')
        self._logger.debug('instructing Agent to start %s', 'transfer server')
        result = self._send_instruction({'process': 'transfer_server_start',
                                         'arguments': {'mode': 'receive',
                                                       'port': host_port}})
        self._logger.debug('Agent responded with: %s', result)
        return result

    def _start_server(self, host_port, process, message):
        """Starts the transfer server."""
        self._logger.debug('instructing Agent to start %s', message)
        result = self._send_instruction({'process': process,
                                         'arguments': {'port': host_port}})
        self._logger.debug('Agent responded with: %s', result)
        return result

    # def start_agent(self):
    #     print("the BootstrapAgent-module is responsible for starting agent.py (Agent) on destination host")

    def stop_agent(self):
        """Starts the Agent on destination host."""
        self._logger.debug('instructing Agent to stop itself')
        result = self._send_instruction({'process': 'stop',
                                         'arguments': {}})
        self._logger.debug('Agent responded with: %s', result)
        return result

    def send_command(self, command):
        """Executes Linux command and returns the response in a byte list."""
        if command.lower().strip() == 'exit':
            response = [b'ABORTED: hit control + c to end interactive mode']
            return response
        try:
            command_json = json.dumps({'command': command})  # serialize dict to a JSON formatted str
            command_byte = command_json.encode('utf-8')  # encode JSON formatted string to byte
            with urllib.request.urlopen(f'http://{self.host}:{self.group_ports["local_port_command"]}',
                                        timeout=5, data=command_byte) as request_obj:
                # response = request_obj.read().decode('utf-8')  # from byte to string
                response = request_obj.read()
        except URLError as exp:
            self._logger.error('instructing the Agent raised an error: %s', exp)
            response = False
        except ConnectionResetError as exp:
            self._logger.error('instructing the Agent raised an error: %s', exp)
            response = False
        except timeout:
            self._logger.error('timeout exceeded when instructing the Agent')
            response = False
        return response

    def send_file(self, src_file_path, dest_path):
        """Opens the socket and connects to transfer server (ie. Agent) and sends a file."""
        port = self.group_ports["local_port_transfer"]
        data_protocol = DataProtocol('send', port)
        data_protocol.start()
        self._logger.debug('connection from client to transfer server by Agent on destination host established')
        result = False
        try:
            result = all([data_protocol.send_metadata(src_file_path, dest_path),
                          data_protocol.send_file(src_file_path)])
        except FileNotFoundError:
            self._logger.error('file or directory does not exist on client')
        data_protocol.stop()
        return result  # lost-exception / return statement in finally block may swallow exception (col 12)

    @abstractmethod
    def start(self):
        """Starts the necessary programs on target destination host."""

    @abstractmethod
    def stop(self):
        """Terminates the started program(s) and the Agent on target destination host."""


class ForInstructor(Instructor):
    """Provides interaction with the Agent, which resides on target destination host, to accommodate FOR mode.

    Functions:
    - forwards connections (FOR) implicitly, because SSH is responsible for forwarding connections, not the Agent
    - interaction with the heartbeat responder
    - provide access to OS services (COMMAND)
    - transfer files (TRANSFER)
    """

    def __init__(self, group_ports):
        """Initializes the ForInstructor object.

        Args:
            group_ports (dict): A group of ports for powermole to bind on (localhost and target destination host)

        """
        Instructor.__init__(self, group_ports)

    def start(self):
        """Starts the heartbeat responder."""
        return all([self._start_heartbeat_responder(host_port=self.group_ports["remote_port_heartbeat"]),
                    self._start_command_server(host_port=self.group_ports["remote_port_command"]),
                    self._start_transfer_server(host_port=self.group_ports["remote_port_transfer"])])

    def stop(self):
        """Terminates the Agent on destination host and started services (heartbeat responder, command server, etc)."""
        return self.stop_agent()


class TorInstructor(Instructor):
    """Provides interaction with the Agent, which resides on target destination host, to accommodate Tor mode.

    Functions:
    - proxify internet traffic (TOR)
    - interaction with the heartbeat responder
    - provide access to OS services (COMMAND)
    - transfer files (TRANSFER)
    """

    def __init__(self, group_ports, ip_address_i='localhost', ip_address_e='localhost'):
        """Initializes the TorInstructor object.

        Args:
            group_ports (dict): A group of ports for powermole to bind on (localhost and target destination host)
            ip_address_i (basestring): The IP address on host for incoming SOCKS encapsulated connections.
            ip_address_e (basestring): The IP address on host (on a possible different ifcard) for outgoing connections.

        """
        Instructor.__init__(self, group_ports)
        self.ip_address_i = ip_address_i
        self.ip_address_e = ip_address_e

    def _start_proxy_server(self, address_i, port_i, address_e):
        """Sends an instruction to Agent to start the proxy server."""
        self._logger.debug('instructing Agent to start the proxy server')
        result = self._send_instruction({'process': 'proxy_server_start',
                                         'arguments': {'address_i': address_i,
                                                       'port_i': port_i,
                                                       'address_e': address_e}})
        self._logger.debug('Agent responded with: %s', result)
        return result

    def start(self):
        """Starts the SOCKS proxy and heartbeat responder."""
        return all([self._start_proxy_server(address_i=self.ip_address_i,
                                             port_i=self.group_ports["remote_port_proxy"],
                                             address_e=self.ip_address_e),
                    self._start_heartbeat_responder(host_port=self.group_ports["remote_port_heartbeat"]),
                    self._start_command_server(host_port=self.group_ports["remote_port_command"]),
                    self._start_transfer_server(host_port=self.group_ports["remote_port_transfer"])])

    def stop(self):
        """Terminates the Agent on destination host and all started services (heartbeat responder, proxy server, ..)."""
        return self.stop_agent()


class PlainInstructor(Instructor):
    """Provides interaction with the Agent, which resides on target destination host, to accommodate FOR mode.

    Functions:
    - interaction with the heartbeat responder
    - provide access to OS services (COMMAND)
    - transfer files (TRANSFER)
    """

    def __init__(self, group_ports):
        """Initializes the ForInstructor object.

        Args:
            group_ports (dict): A group of ports for powermole to bind on (localhost and target destination host)

        """
        Instructor.__init__(self, group_ports)

    def start(self):
        """Starts the heartbeat responder."""
        return all([self._start_heartbeat_responder(host_port=self.group_ports["remote_port_heartbeat"]),
                    self._start_command_server(host_port=self.group_ports["remote_port_command"]),
                    self._start_transfer_server(host_port=self.group_ports["remote_port_transfer"])])

    def stop(self):
        """Terminates the Agent on destination host and started services (heartbeat responder, command server, etc)."""
        return self.stop_agent()
